# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from six import text_type as str
from six import text_type as str
from six.moves import urllib
from six.moves.html_parser import HTMLParser
from zipfile import ZipFile
from json import JSONEncoder, JSONDecoder

import os
import copy
import uuid
import math
import inspect
import argparse
import unicodedata

__author__ = 'dmiro'
__version_info__ = (1, 0, 3)
__version__ = '.'.join(str(v) for v in __version_info__)



class BagOfWords(object):
    """Implementing a bag of words with their frequency of usages"""

    def __init__(self, *args):
        self._bow = {}
        self.add(*args)

    def __calc(self, operation, *args):
        for words in args:
            if isinstance(words, str):
                words = [words]
            for word in words:
                n = 1
                if isinstance(words, dict):
                    n = words[word]
                self._bow[word] = operation(self._bow.get(word, 0), n)
                if self._bow[word] < 1:
                    del self._bow[word]

    def add(self, *args):
        """Add set of word, word list or word dict to bag of words.
        :param args: set of word or word list to add
        :return:nothing
        """
        self.__calc(lambda x,y: x+y, *args)

    def delete(self, *args):
        """Delete set of word, word list or word dict to bag of words.
        :param args: set of word or word list to add
        :return:nothing
        """
        self.__calc(lambda x,y: x-y, *args)

    @property
    def rates(self):
        """Rate of occurrences
        :return: Dict
        """
        total = float(self.num())
        if total:
            return {k:v/total for k, v in list(self._bow.items())}
        else:
            return {}

    @property
    def sorted_rates(self):
        """Sorted rate of occurrences
        :return: list sorted from greater to lowest rate
        """
        total = float(self.num())
        if total:
            res = [(k,v/total) for k, v in list(self._bow.items())]
            return sorted(res, key=lambda t: t[1], reverse=True)
        else:
            return []

    def freq(self, word):
        """Frequency of a word.
        :param word: word to query
        :return: frequency
        """
        if word in self._bow:
            return self._bow[word]
        else:
            return 0

    def rate(self, word):
        """Rate of a word.
        :param word: word to query
        :return: rate
        """
        total = float(self.num())
        if total:
            return self.freq(word)/total
        else:
            return 0

    def __add__(self, other):
        """ Overloading of "+" operator to join BagOfWord+BagOfWord, BagOfWords+str or
        BagOfWords+list.
        :param other: BagOfWords, str or list
        :return: BagOfWords
        """
        result = self.copy()
        if isinstance(other, BagOfWords):
            result.add(dict(other))
        else:
            result.add(other)
        return result

    def __sub__(self, other):
        """ Overloading of "-" operator to join BagOfWord+BagOfWord, BagOfWords+str or
        BagOfWords+list.
        :param other: BagOfWords, str or list
        :return: BagOfWords
        """
        result = self.copy()
        if isinstance(other, BagOfWords):
            result.delete(dict(other))
        else:
            result.delete(other)
        return result

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __iter__(self):
        return list(self._bow.items())

    def __getitem__(self, offset):
        return self._bow.__getitem__(offset)

    def __len__(self):
        return self._bow.__len__()

    def __repr__(self):
        return self._bow.__repr__()

    def __delitem__(self, key):
        del self._bow[key]

    def __eq__(self, other):
        if isinstance(other, BagOfWords):
            return self._bow == other._bow
        else:
            return self._bow == other

    def __ne__(self, other):
        if isinstance(other, BagOfWords):
            return self._bow !=other._bow
        else:
            return self._bow != other


    def copy(self):
        return copy.deepcopy(self)

    def clear(self):
        """Clear word list."""
        self._bow.clear()

    def items(self):
        """Return an iterator over the word dictionary’s (word, frequency) pairs."""
        return list(self._bow.items())

    def keys(self):
        """Word list contained in the object."""
        return list(self._bow.keys())

    def words(self):
        """Word list contained in the object."""
        return list(self.keys())

    def items(self):
        return list(self._bow.items())

    def values(self):
        return list(self._bow.values())

    def num(self):
        """Total number of words."""
        return sum(self._bow.values())

    def __contains__(self, key):
        """Method key in y"""
        return key in self._bow

    def __call__(self, *args):
         self.add(self, *args)


class TextFilters(object):
    """Filters for transforming a text"""

    @staticmethod
    def upper(text):
        """Convert text to uppercase."""
        return text.upper()

    @staticmethod
    def lower(text):
        """Convert text to lowercase."""
        return text.lower()

    @staticmethod
    def invalid_chars(text):
        """Remove invalid chars from a text."""
        INVALID_CHARS = "/\¨º-~#@|¡!,·$%&()¿?'[^""`]+}{><;,:.=*^_"
        return ''.join([char for char in text if char not in INVALID_CHARS])

    @staticmethod
    def html_to_text(text):
        """Conversion from HTML markup to plain text."""
        class _HTMLParser(HTMLParser):

            def __init__(self):
                HTMLParser.__init__(self)
                self.text = []

            def handle_data(self, data):
                append = True
                text = data.split()
                if text:
                    tag = self.get_starttag_text()
                    if tag:
                        tag = tag.lower()
                        append = not tag.startswith(('<script','<style'))
                    if append:
                        self.text.extend(text)
        parser = _HTMLParser()
        parser.feed(text)
        return ' '.join(parser.text)


class WordFilters(object):
    """Filters for transforming a set of words"""

    @staticmethod
    def stemming(lang, stemming, words):
        """Lemmatize text.
        :param lang: lang text to lemmatize
        :param stemming: number loops of lemmatizing
        """
        import Stemmer as stemmer
        try:
            stemmer = stemmer.Stemmer(lang)
            for i in range(stemming):
                words = stemmer.stemWords(words)
            return words
        except KeyError:
            return words

    @staticmethod
    def stopwords(lang, words):
        """Remove stop words from a text.
        :param lang: language text where remove empty words
        """
        import stop_words
        try:
            stopwords = stop_words.get_stop_words(lang)
            return [word for word in words if word not in stopwords]
        except stop_words.StopWordError:
            return words

    @staticmethod
    def normalize(words):
        """Normalize chars from a text."""
        return [''.join((char for char in unicodedata.normalize('NFD', str(word)) if unicodedata.category(char) != 'Mn'))
                for word in words]


class Tokenizer(object):
    """Allows to break a string into tokens (set of words). Optionally allows you to set
    filters before (TextFilters) and after (WordFilters) breaking the string into tokens.
    """

    def __init__(self):
        object.__init__(self)

    def before_tokenizer(self, textfilters, text):
        """function to call before tokenizer text.
        :param textfilters: static class with helper methods to filter text
        :param text: The text will be split
        """
        return text

    def after_tokenizer(self, wordfilters, words):
        """function to call after tokenizer text.
        :param wordfilters: static class with helper methods to filter words
        :param words: split words
        """
        return words

    def tokenizer(self, text):
        """tokenize a text.
        :param text: text to tokenizer
        """
        text = self.before_tokenizer(TextFilters, text)
        words = text.split()
        words = self.after_tokenizer(WordFilters, words)
        return words

    def __call__(self, text):
        return self.tokenizer(text)


class Document(BagOfWords, Tokenizer):
    """Implementing a bag of words where all words are of the same category. Retrieves
    the text of a file, folder, url or zip, and also allows save or retrieve the Document
    in json format.
    """

    def __init__(self):
        Tokenizer.__init__(self)
        BagOfWords.__init__(self)
        self.numdocs = 0

    def _read(self, id_, text):
        self.numdocs += 1
        words = self.tokenizer(text)
        self.add(words)

    def clear(self):
        """Clear word list."""
        BagOfWords.clear(self)
        self.numdocs = 0

    def read_text(self, text):
        """The text is stored in a BagOfWords identified by Id.
        :param text: text to add a BagOfWords
        :return: nothing
        """
        self._read(None, text)

    def read_files(self, *filenames):
        """The contents of each file or files is stored in a BagOfWord identified by the
        filename.
        :param *filenames: filenames to add
        :return: nothing
        """
        for filename in filenames:
            text = open(filename, 'r').read()
            self._read(filename, text)

    def read_dir(self, *paths):
        """The contents of each file o files of a directory is stored in a BagOfWord
        identified by the filename.
        :param paths: directory or directories path to add files
        :return: nothing
        """
        for path in paths:
            fn = []
            for (_, _, filenames) in os.walk(path):
                fn.extend([os.path.join(path,f) for f in filenames])
                break
            self.read_files(*fn)

    def read_urls(self, *urls):
        """The contents of each url or urls is stored in a BagOfWord identified by the url.
        :param *urls: urls to add
        :return: nothing
        """
        headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; rv:24.0) Gecko/20140129 Firefox/24.0'}
        for url in urls:
            req = urllib.Request(url=url, headers=headers)
            text = urllib.request.urlopen(req).read()
            self._read(url, text)

    def read_zips(self, *zipfilenames):
        """The contents of each file o files of a zip file is stored in a BagOfWord
        identified by the filename.
        :param *zipfilenames: zip files to add
        :return: nothing
        """
        for zipfilename in zipfilenames:
            input_zip = ZipFile(zipfilename)
            for input_file in input_zip.infolist():
                if input_file.file_size > 0:
                    text = input_zip.read(input_file)
                    self._read(input_file.filename, text)

    def to_json(self):
        """Convert Document object to json string.
        :return: json string
        """
        class _Encoder(JSONEncoder):

            def default(self, obj):
                if isinstance(obj, DocumentClass) or \
                   isinstance(obj, BagOfWords):
                    d = {'__class__': obj.__class__.__name__,
                         '__module__':obj.__module__}
                    d.update(obj.__dict__)
                    return d
                if not inspect.isfunction(obj):
                    return super(_Encoder, self).default(obj)

        return _Encoder().encode(self)

    @staticmethod
    def from_json(json_):
        """Convert json string to Document object.
        :param json_: json string
        :return: Document object
        """
        class _Decoder(JSONDecoder):

            def __init__(self):
                JSONDecoder.__init__(self, object_hook=self.dict_to_object)

            def dict_to_object(self, d):
                if '__class__' in d:
                    class_name = d.pop('__class__')
                    module_name = d.pop('__module__')
                    module = __import__(module_name)
                    class_ = getattr(module, class_name)
##                    if issubclass(class_, BagOfWords):
##                        obj = class_(d.pop('_bow'))
##                    else:
##                        obj = class_()
                    obj = class_()
                    for k, v in list(d.items()):
                        setattr(obj, k, v)
                    return obj
                return d

        return _Decoder().decode(json_)

    def save(self, filename):
        """Serialize Documentand save to a file in json format
        :filename: file to save
        :return: nothing
        """
        with open(filename, 'w') as f:
            json_ = self.to_json()
            f.write(json_)

    @staticmethod
    def load(filename):
        """Load and deserialize Document from file saved in json format
        :filename: file to load
        :return: nothing
        """
        with open(filename, 'r') as f:
            json_ = f.read()
            return Document.from_json(json_)

    def __call__(self, text):
        self.read_text(text)


class DocumentClass(Document):
    """Implementing a bag of words collection where all the bags of words are the same
    category, as well as a bag of words with the entire collection of words. Each bag
    of words has an identifier otherwise it's assigned an calculated identifier.
    Retrieves the text of a file, folder, url or zip, and also allows save or retrieve
    the collection in json format.
    """

    def __init__(self):
        Document.__init__(self)
        self.docs = {}

    def _read(self, id_, text):
        words = self.tokenizer(text)
        bow = BagOfWords(words)
        if not id_:
            id_ = uuid.uuid4().hex
        if id_ in self.docs:
            self.delete(dict(self.docs[id_]))
        else:
            self.numdocs += 1
        self.docs[id_] = bow
        self.add(words)

    def clear(self):
        """Clear word and docs list."""
        Document.clear(self)
        self.docs = {}

    def read_text(self, text, id_=None):
        """The text is stored in a BagOfWords identified by Id.
        :param text: text to add a BagOfWords
        :param id_: BagOfWord identifier. Optional. If not set then it's set an UUID4
        identifier.
        :return: nothing
        """
        self._read(id_, text)

    def __call__(self, text, id_=None):
        self._read(id_, text)


class DefaultTokenizer(Tokenizer):
    """Tokenizer subclass that implements the text filters 'lower', 'invalid_chars'
    and the word filters 'stopwords', 'stemming' and 'normalize'.
    """

    def __init__(self, lang='english', stemming=1):
         Tokenizer.__init__(self)
         self.lang = lang
         self.stemming = stemming

    def before_tokenizer(self, textfilters, text):
        text = textfilters.lower(text)
        text = textfilters.invalid_chars(text)
        return text

    def after_tokenizer(self, wordfilters, words):
        words = wordfilters.stopwords(self.lang, words)
        words = wordfilters.stemming(self.lang, self.stemming, words)
        words = wordfilters.normalize(words)
        return words


class SimpleTokenizer(Tokenizer):
    """Tokenizer subclass that implements the text filters 'lower', 'invalid_chars'
    and the word filter 'normalize'.
    """

    def __init__(self):
         Tokenizer.__init__(self)

    def before_tokenizer(self, textfilters, text):
        text = textfilters.lower(text)
        text = textfilters.invalid_chars(text)
        return text

    def after_tokenizer(self, wordfilters, words):
        words = wordfilters.normalize(words)
        return words


class HtmlTokenizer(DefaultTokenizer):
    """Tokenizer subclass that implements the text filters 'htm_to_text', 'lower',
    'invalid_chars' and the word filter 'normalize'.
    """

    def __init__(self, lang='english', stemming=1):
         DefaultTokenizer.__init__(self, lang, stemming)

    def before_tokenizer(self, textfilters, text):
        text = textfilters.html_to_text(text)
        text = DefaultTokenizer.before_tokenizer(self, textfilters, text)
        return text


class DefaultDocument(Document, DefaultTokenizer):
    """DefaultTokenizer and Document subclass"""

    def __init__(self, lang='english', stemming=1):
        Document.__init__(self)
        DefaultTokenizer.__init__(self, lang, stemming)


class SimpleDocument(Document, SimpleTokenizer):
    """SimpleTokenizer and Document subclass"""

    def __init__(self):
        Document.__init__(self)
        SimpleTokenizer.__init__(self)


class HtmlDocument(Document, HtmlTokenizer):
    """HtmlTokenizer and Document subclass"""

    def __init__(self, lang='english', stemming=1):
        Document.__init__(self)
        HtmlTokenizer.__init__(self, lang, stemming)


class DefaultDocumentClass(DocumentClass, DefaultTokenizer):
    """DefaultTokenizer and DocumentClass subclass"""

    def __init__(self, lang='english', stemming=1):
        DocumentClass.__init__(self)
        DefaultTokenizer.__init__(self, lang, stemming)


class SimpleDocumentClass(DocumentClass, SimpleTokenizer):
    """SimpleTokenizer and DocumentClass subclass"""

    def __init__(self):
        DocumentClass.__init__(self)
        SimpleTokenizer.__init__(self)


class HtmlDocumentClass(DocumentClass, HtmlTokenizer):
    """HtmlTokenizer and DocumentClass subclass"""

    def __init__(self, lang='english', stemming=1):
        DocumentClass.__init__(self)
        HtmlTokenizer.__init__(self, lang, stemming)


def document_classifier(document, **classifieds):
    """Text classification based on an implementation of Naive Bayes
    :param document: document class instance to classify.
    :param classifieds: dictionary with Document class instances have already been classified.
    :return: list sorted from highest to lowest probability.
    """
    # http://blog.yhathq.com/posts/naive-bayes-in-python.html
    res = {}
    total_docs = SimpleDocument()
    for classified in list(classifieds.values()):
        total_docs += classified
    for k_classified, classified in list(classifieds.items()):
        prior = float(classified.num()) / float(total_docs.num())
        log_prob = 0.0
        for word, value in list(document.items()):
            if word in total_docs:
                if classified.rate(word) > 0.0:
                    # log(probability) it requires fewer decimal places
                    log_prob += math.log(value * classified.rate(word) / total_docs.rate(word))
        # log space to regular space
        exp_prob = math.exp(log_prob + math.log(prior))
        res[k_classified] = exp_prob
    total = sum(res.values())
    res = [(k,v/total) for k, v in list(res.items())]
    return sorted(res, key=lambda t: t[1], reverse=True)


def _show_document(document, filename, verbose, top=50):
    print('* filename: %s' % filename)
    print('* filter:')
    print('    type: %s' % document.__class__.__name__)
    print('    lang: %s' % document.lang)
    print('    stemming: %s' % document.stemming)
    print('* total words: %d' % document.num())
    print('* total docs: %d' % document.numdocs)
    if verbose:
        if top:
            words = 'word (top %d)' % top
            rates = document.sorted_rates[0:top]
        else:
            words = 'word'
            rates = document.sorted_rates
        posadj = len(str(len(rates)))+1
        print('*','pos'.rjust(posadj),'|',words.ljust(35),'|','occurrence'.rjust(10),\
              '|','rate'.rjust(10))
        print(' ','-'*posadj,'|','-'*35,'|','-'*10,'|','-'*10)
        for word, rate in rates:
            print(' ',str(rates.index((word, rate))+1).rjust(posadj),'|',\
                  word.encode('utf-8').ljust(35),'|', str(document[word]).rjust(10),\
                  '|',('%.8f' % rate).rjust(10))


def _show(args):
    try:
        dc = Document.load(args.filename)
        _show_document(document=dc, filename=args.filename, verbose=True, top=args.list_top_words)
    except IOError:
        print('No such classifier: %s' % args.filename)


def _create(args):
    if args.filter == 'html':
        dc = HtmlDocument(lang=args.lang_filter, stemming=args.stemming_filter)
    else:
        dc = DefaultDocument(lang=args.lang_filter, stemming=args.stemming_filter)
    dc.save(args.filename)
    _show_document(document=dc, filename=args.filename, verbose=False)


def _learn(args):
    try:
        dc = Document.load(args.filename)
        if args.rewrite:
            dc.clear()
        print('\ncurrent')
        print('=======')
        _show_document(document=dc, filename=args.filename, verbose=False)
        print('\nupdated')
        print('=======')
        if args.url:
            dc.read_urls(*args.url)
        if args.dir:
            dc.read_dir(*args.dir)
        if args.file:
            dc.read_files(*args.file)
        if args.zip:
            dc.read_zips(*args.zip)
        if not args.no_learn:
            dc.save(args.filename)
        _show_document(document=dc, filename=args.filename, verbose=True, top=args.list_top_words)
    except IOError:
        print('No such classifier: %s' % args.filename)


def _classify(args):
    dclist = {}
    for filename in args.classifiers:
        dc = Document.load(filename)
        dclist[filename] = dc
    dc = list(dclist.values())[0].copy()
    dc.clear()
##    if args.filter == 'html':
##        dc = HtmlDocument(lang=args.lang_filter, stemming=args.stemming_filter)
##    else:
##        dc = DefaultDocument(lang=args.lang_filter, stemming=args.stemming_filter)
    if args.text:
        dc.read_text(args.text)
    elif args.url:
        dc.read_urls(args.url)
    elif args.file:
        dc.read_files(args.file)
    result = document_classifier(dc, **dclist)
    print('*','classifier'.ljust(35),'|','rate'.rjust(10))
    print(' ','-'*35,'|','-'*10)
    for classifier, rate in result:
        print(' ',classifier.encode('utf-8').ljust(35),'|',('%.8f' % rate).rjust(10))


def main():
    parser = argparse.ArgumentParser(description='Manage several document to apply text classification.',
                                     epilog="see https://github.com/dmiro/bagofwords for more info")
    parser.add_argument('--version', action='version', version=__version__,
                        help='show version and exit')
    subparsers = parser.add_subparsers(help='')
    # create command
    parser_create = subparsers.add_parser('create', help='create classifier')
    parser_create.add_argument('filter', choices=['text', 'html'], help='filter type')
    parser_create.add_argument('filename', help='file to be created where words learned are saved')
    parser_create.add_argument('--lang-filter', default='english', type=str,
                               help='language text where remove empty words')
    parser_create.add_argument('--stemming-filter', default=1, type=int,
                               help='number loops of lemmatizing')
    parser_create.set_defaults(func=_create)
    # learn command
    parser_learn = subparsers.add_parser('learn', help='add words learned a classifier')
    parser_learn.add_argument('filename', help='file to write words learned')
    parser_learn.add_argument('--file', nargs='+', help='filenames to learn')
    parser_learn.add_argument('--dir', nargs='+', help='directories to learn')
    parser_learn.add_argument('--url', nargs='+', help='url resources to learn')
    parser_learn.add_argument('--zip', nargs='+', help='zip filenames to learn')
    parser_learn.add_argument('--no-learn', action='store_true', default=False,
                              help='not write to file the words learned')
    parser_learn.add_argument('--rewrite', action='store_true', default=False,
                              help='overwrite the file')
    parser_learn.add_argument('--list-top-words', default=50, type=int,
                              help='maximum number of words to list, 50 by default, -1 list all')
    parser_learn.set_defaults(func=_learn)
    # show command
    parser_show = subparsers.add_parser('show', help='show classifier info')
    parser_show.add_argument('filename', help='filename')
    parser_show.add_argument('--list-top-words', default=50, type=int,
                             help='maximum number of words to list, 50 by default, -1 list all')
    parser_show.set_defaults(func=_show)
    # classify command
    parser_classify = subparsers.add_parser('classify', help='Naive Bayes text classification')
    parser_classify.add_argument('classifiers', nargs='+', help='classifiers')
    parser_classify.add_argument('--file', help='file to classify')
    parser_classify.add_argument('--url', help='url resource to classify')
    parser_classify.add_argument('--text',help='text to classify')
    parser_classify.set_defaults(func=_classify)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
	main()



from setuptools import setup, find_packages

PROJECT = "bagofwords"

long_description = ''

try:
    import subprocess
    import pandoc

    process = subprocess.Popen(
        ['which pandoc'],
        shell=True,
        stdout=subprocess.PIPE,
        universal_newlines=True
    )

    pandoc_path = process.communicate()[0]
    pandoc_path = pandoc_path.strip('\n')

    pandoc.core.PANDOC_PATH = pandoc_path

    doc = pandoc.Document()
    doc.markdown = open('README.md').read()

    long_description = doc.rst

except:
    pass

setup(
    name=PROJECT,
    version=__import__("bow").__version__,
    author = "David Miro <lite.3engine@gmail.com>",
    author_email = 'lite.3engine@gmail.com',
    description = "The main goal this Python module is to provide functions to apply Text Classification.",
    long_description=long_description,
    license=open('LICENSE').read(),
    url='https://github.com/dmiro/bagofwords',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License'
        ],
    py_modules=['bow'],
    entry_points = {
        'console_scripts': ['bow = bow:main']
        },
    install_requires=[
        'stop-words',
        'PyStemmer',
        'six'
        ],
    test_suite = 'test',
    platforms=['Any'],
    zip_safe=False
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import unittest
from unittest import TestCase
import bow
from bow import BagOfWords, TextFilters, WordFilters, Tokenizer, SimpleTokenizer, DefaultTokenizer, DocumentClass, DefaultDocumentClass, DefaultDocument, SimpleDocument
import mock
import six


if six.PY2:
    TestCase.assertCountEqual = TestCase.assertEqual

        
class BagOfWordsTest(TestCase):
    def __init__(self, *args, **kwargs):
        super(BagOfWordsTest, self).__init__(*args, **kwargs)
    
    def setUp(self):
        self.bow = BagOfWords()

    def test_add_one_word(self):
        self.bow.add('David')
        self.bow.add({'David':2})
        self.assertCountEqual(self.bow.words(), ['David'])
        self.assertEqual(len(self.bow), 1)
        self.assertEqual(self.bow.num(), 3)
        self.assertEqual(self.bow.freq('David'), 3)
        self.assertCountEqual(dict(self.bow), {'David':3})

    def test_add_two_words(self):
        self.bow.add('David', ['David','Álex'])
        self.assertCountEqual(self.bow.words(), ['Álex', 'David'])
        self.assertEqual(len(self.bow), 2)
        self.assertEqual(self.bow.num(), 3)
        self.assertEqual(self.bow.freq('David'), 2)
        self.assertCountEqual(dict(self.bow), {'Álex':1, 'David':2})

    def test_del_one_word(self):
        self.bow.delete('David')
        self.assertCountEqual(dict(self.bow), {})
        #
        self.bow.add('David')
        self.bow.delete('David')
        self.assertCountEqual(dict(self.bow), {})
        #
        self.bow.add('David', 'David')
        self.bow.delete('David')
        self.assertCountEqual(self.bow.words(), ['David'])
        self.assertEqual(len(self.bow), 1)
        self.assertEqual(self.bow.num(), 1)
        self.assertEqual(self.bow.freq('David'), 1)
        self.assertCountEqual(dict(self.bow), {'David':1})

    def test_del_two_word(self):
        self.bow.delete('David', 'Álex')
        self.assertCountEqual(dict(self.bow), {})
        #
        self.bow.add('David', 'Álex')
        self.bow.delete('David', 'Álex')
        self.assertCountEqual(dict(self.bow), {})
        #
        self.bow.add({'David':2})
        self.bow.delete('David')
        self.bow.add('Álex')
        self.assertCountEqual(self.bow.words(), ['Álex', 'David'])
        self.assertEqual(len(self.bow), 2)
        self.assertEqual(self.bow.num(), 2)
        self.assertEqual(self.bow.freq('David'), 1)
        self.assertCountEqual(dict(self.bow), {'Álex':1, 'David':1})

    def test_join_add(self):
        a = BagOfWords('car', 'chair', 'chicken')
        b = BagOfWords({'chicken':2}, ['eye', 'ugly'])
        c = BagOfWords('plane')
        self.assertCountEqual(dict(a + b + c), {'car': 1, 'chair': 1, 'eye': 1, 'chicken': 3, 'plane': 1, 'ugly': 1})
        self.assertCountEqual(dict(c + b + a), {'car': 1, 'chair': 1, 'eye': 1, 'chicken': 3, 'plane': 1, 'ugly': 1})
        self.assertCountEqual(dict(b + c + a), {'car': 1, 'chair': 1, 'eye': 1, 'chicken': 3, 'plane': 1, 'ugly': 1})
        #
        total = a + b + c
        total = 'ugly' + total
        self.assertCountEqual(dict(total), {'car': 1, 'chair': 1, 'eye': 1, 'chicken': 3, 'plane': 1, 'ugly': 2})
        #
        total = a + b + c
        total = 'ugly' + total
        total = total + 'plane'
        self.assertCountEqual(dict(total), {'car': 1, 'chair': 1, 'eye': 1, 'chicken': 3, 'plane': 2, 'ugly': 2})
        #
        total = a + b + c
        total = total + ['car', 'chair', 'chicken'] + ['chicken', 'chicken', 'eye']
        self.assertCountEqual(dict(total), {'car': 2, 'chair': 2, 'eye': 2, 'chicken': 6, 'plane': 1, 'ugly': 1})

    def test_join_sub(self):
        a = BagOfWords('car', 'chair', 'chicken')
        b = BagOfWords({'chicken':2}, ['eye', 'ugly'])
        c = BagOfWords('plane')
        self.assertCountEqual(dict(a - b - c), {'car': 1, 'chair': 1})
        self.assertCountEqual(dict(c - b - a), {'plane': 1})
        self.assertCountEqual(dict(b - c - a), {'chicken':1, 'eye':1, 'ugly':1})
        #
        total = b - c - a
        total = 'eye' - total
        self.assertCountEqual(dict(total), {'chicken':1, 'ugly':1})
        #
        total = b - c - a
        total = 'eye' - total
        total = total - 'eye'
        self.assertCountEqual(dict(total), {'chicken':1, 'ugly':1})
        #
        total = b - c - a
        total = total - ['chicken', 'ugly']
        self.assertCountEqual(dict(total), {'eye':1})

    def test_clear(self):
        self.bow.add('item', 'item')
        self.bow.clear()
        self.assertEqual(len(self.bow), 0)
        self.assertEqual(self.bow.num(), 0)
        self.assertEqual(self.bow.freq('item'), 0)
        self.assertCountEqual(dict(self.bow), {})

    def test_item(self):
        self.bow.add('item1', 'item2', 'item2', 'item3')
        self.assertEqual(self.bow['item2'], 2)
        self.assertEqual(self.bow['item3'], 1)
        self.assertEqual(self.bow['item1'], 1)

    def test_copy(self):
        a = BagOfWords('car', 'chair', 'chicken')
        b = a.copy()
        self.assertEqual(a == b, True)

    def test_del(self):
        self.bow.add(['car', 'chair', 'chicken'])
        del self.bow['car']
        self.assertCountEqual(dict(self.bow), {'chair':1, 'chicken':1})

    def test_cmp(self):
        a = BagOfWords('car', 'chair', 'chicken')
        b = BagOfWords('car', 'chair', 'chicken')
        self.assertEqual(a == b, True)
        #
        a.add('car')
        self.assertEqual(a == b, False)

    def test_has_key(self):
        self.bow.add('car', 'chair', 'chicken')
        self.assertEqual('car' in self.bow, True)
        self.assertEqual('car' in self.bow, True)

    def test_rate(self):
        self.bow.add(['b','a','a','a'])
        self.assertCountEqual(self.bow.rates, {'a':0.75, 'b':0.25})
        self.assertCountEqual(self.bow.sorted_rates, [('a', 0.75), ('b', 0.25)])  
        self.assertEqual(self.bow.rate('a'), 0.75)
        self.assertEqual(self.bow.rate('b'), 0.25)
        self.assertEqual(self.bow.rate('c'), 0)
        #
        self.bow.clear()
        self.assertEqual(self.bow.rate('a'), 0)


class TokenizerTest(TestCase):

    def test_default_tokenizer(self):
        tokens = DefaultTokenizer()
        words = tokens('How do you convert a tuple to a list?');
        self.assertCountEqual(words, ['convert', 'tupl', 'list'])
        #
        words = tokens.tokenizer('How do you convert a tuple to a list?');
        self.assertCountEqual(words, ['convert', 'tupl', 'list'])
        #
        tokens = DefaultTokenizer(stemming=0)
        words = tokens('How do you convert a tuple to a list?');
        self.assertCountEqual(words, ['convert', 'tuple', 'list'])
        #
        tokens = DefaultTokenizer(lang='', stemming=0)
        words = tokens('How do you convert a tuple to a list?');
        self.assertCountEqual(words, ['how', 'do', 'you', 'convert', 'a', 'tuple', 'to', 'a', 'list'])
        #
        tokens = DefaultTokenizer(lang='spanish')
        words = tokens('Cómo convertir una tupla a lista?');
        self.assertCountEqual(words, ['com', 'convert', 'tupl', 'list'])
        #
        tokens = DefaultTokenizer(lang='spanish', stemming=0)
        words = tokens('Cómo convertir una tupla a lista?');
        self.assertCountEqual(words, ['como', 'convertir', 'tupla', 'lista'])
        #
        tokens = DefaultTokenizer(lang='', stemming=0)
        words = tokens('Cómo convertir una tupla a lista?');
        self.assertCountEqual(words, ['como', 'convertir', 'una', 'tupla', 'a', 'lista'])

    def test_simple_tokenizer(self):
        tokens = SimpleTokenizer()
        words = tokens('How, do you convert - a tuple to a list?');
        self.assertCountEqual(words, ['how', 'do', 'you', 'convert', 'a', 'tuple', 'to', 'a', 'list'])

    def test_tokenizer(self):

        class _MyTokenizer(Tokenizer):

            def __init__(self):
                 Tokenizer.__init__(self)

            def before_tokenizer(self, textfilters, text):
                text = textfilters.upper(text)
                return text

            def after_tokenizer(self, wordfilters, words):
                words = wordfilters.normalize(words)
                return words
        tokens = _MyTokenizer()
        words = tokens('How, do you convert - a tuple to a list?');
        self.assertCountEqual(words, ['HOW,', 'DO', 'YOU', 'CONVERT', '-', 'A', 'TUPLE', 'TO', 'A', 'LIST?'])
        #
        class _MyTokenizer(Tokenizer):

            def __init__(self):
                 Tokenizer.__init__(self)

            def before_tokenizer(self, textfilters, text):
                text = textfilters.html_to_text(text)
                text = textfilters.invalid_chars(text)
                text = textfilters.lower(text)
                return text

            def after_tokenizer(self, wordfilters, words):
                words = wordfilters.stopwords('english', words)
                words = wordfilters.normalize(words)
                return words
        tokens = _MyTokenizer()
        text = '''
                <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
                "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
                <html>
                    <body>
                        <!--my comment -->
                        <script>my script</script>
                        <SCRIPT>another script</script>
                        <style>my style</style>
                        <h3>my project!!</h3> <br>
                        <b>Description</b>:<br/>
                        This small script is intended to allow conversion from HTML markup to plain text.
                    </body>
                </html>
                '''
        words = tokens(text)
        self.assertCountEqual(words, ['project', 'description', 'small', 'script', 'intended', 'allow', 'conversion',
                                 'html', 'markup', 'plain', 'text'])


class DocumentClassTest(TestCase):

    def test_default_document_class(self):
        dclass = DefaultDocumentClass()
        dclass('hello a beautiful world!', 'text one')
        dclass('hello the Moon!', 'text two')
        dclass('hello the world!', 'text one')
        self.assertCountEqual(dclass.docs, {'text two': {'hello': 1, 'moon': 1}, 'text one': {'world': 1, 'hello': 1}})
        self.assertEqual(dclass, {'world': 1, 'hello': 2, 'moon': 1})
        self.assertEqual(dclass.numdocs, 2)

    def test_default_document(self):
        dclass = DefaultDocument()
        dclass('hello a beautiful world!')
        dclass('hello the Moon!')
        dclass('hello the world!')
        self.assertEqual(dclass, {'world': 2, 'hello': 3, 'beauti': 1, 'moon': 1})
        self.assertEqual(dclass.numdocs, 3)

    def test_json(self):
        dclass = DefaultDocumentClass(lang='spanish')
        dclass.read_text('Hola mundo!', id_='1')
        dclass.read_text('Este es un bonito mundo', id_='2')
        json_ = dclass.to_json()
        dclass = DocumentClass.from_json(json_)
        self.assertCountEqual(dclass.__class__.__name__ , 'DefaultDocumentClass')
        self.assertCountEqual(dclass.docs, {'2': {'mund': 1, 'bonit': 1}, '1': {'mund': 1, 'hol': 1}})
        self.assertEqual(dclass, {'mund': 2, 'hol': 1, 'bonit': 1})
        self.assertEqual(dclass.numdocs, 2)
        self.assertEqual(dclass.lang, 'spanish')
        self.assertEqual(dclass.stemming, 1)

class DocumentClassifierTest(TestCase):

    def test_simple(self):
        docnumbers = bow.SimpleDocument()
        docnumbers('one two three four')
        docnumbers('five six seven')
        docanimals = bow.SimpleDocument()
        docanimals('dog cat')
        docanimals('horse frog')
        docanimals('dog cat')
        docanimals('dog cat')
        docanimals('dog cat')
        docvehicles = bow.SimpleDocument()
        docvehicles('truck car')
        doc = bow.SimpleDocument()
        doc('I am a cat')
        result = bow.document_classifier(doc, numbers=docnumbers, animals=docanimals, vehicles=docvehicles)
        self.assertCountEqual(result, [('animals', 0.6785714285714286), ('numbers', 0.25), ('vehicles', 0.07142857142857142)])
        doc.clear()
        doc('one dog, one cat, three trucks')
        result = bow.document_classifier(doc, numbers=docnumbers, animals=docanimals, vehicles=docvehicles)
        self.assertCountEqual(result, [('numbers', 0.7302518458581976), ('animals', 0.2555881460503691), ('vehicles', 0.014160008091433189)])

    def test_save_document(self):
        if six.PY3:
            # skip this test if python 3
            return
        m = mock.mock_open()
        with mock.patch('bow.open', m, create=True): 
            docnumbers = bow.SimpleDocument()
            docnumbers('one two three four')
            docnumbers('one two three')
            docnumbers.save('test.dat')
        # print(m.mock_calls)
        m.assert_called_once_with('test.dat','w')
        handle = m()
        data = '{"__module__": "bow", "numdocs": 2, "__class__": "SimpleDocument", "_bow": {"four": 1, "three": 2, "two": 2, "one": 2}}'
        handle.write.assert_called_once_with(data)

    def test_load_document(self):
        m = mock.mock_open()
        data = '{"__module__": "bow", "numdocs": 2, "__class__": "SimpleDocument", "_bow": {"four": 1, "three": 2, "two": 2, "one": 2}}'
        with mock.patch('bow.open', mock.mock_open(read_data=data), create=True) as m:
            docnumbers = SimpleDocument.load('test.dat')
        m.assert_called_once_with('test.dat','r')
        self.assertEqual(docnumbers, {'four': 1, 'one': 2, 'three': 2, 'two': 2})
        
       
if __name__ == '__main__':
    unittest.main()
